# ----------------------------------------------------------------------
# |
# |  EntryPoint.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-24 20:13:01
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Automatically generates semantic versions based on changes in a git repository."""

import sys

from io import StringIO
from pathlib import Path
from typing import Annotated, Callable, Optional

import typer

from dbrownell_Common.ContextlibEx import ExitStack  # type: ignore [import-untyped]
from dbrownell_Common.Streams.DoneManager import DoneManager, Flags as DoneManagerFlags  # type: ignore [import-untyped]
from dbrownell_Common.Streams.StreamDecorator import TextWriterT  # type: ignore [import-untyped]
from typer.core import TyperGroup  # type: ignore [import-untyped]

from AutoGitSemVer import (
    GenerateStyle,
    GetSemanticVersion,
    GetSemanticVersionResult,
    __version__,
)


# ----------------------------------------------------------------------
class NaturalOrderGrouper(TyperGroup):
    # pylint: disable=missing-class-docstring
    # ----------------------------------------------------------------------
    def list_commands(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.commands.keys()


# ----------------------------------------------------------------------
app = typer.Typer(
    cls=NaturalOrderGrouper,
    help=__doc__,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


# ----------------------------------------------------------------------
@app.command(
    "Generate",
    help=__doc__,
    no_args_is_help=False,
)
def Generate(
    path: Annotated[
        Path,
        typer.Argument(
            file_okay=False,
            exists=True,
            resolve_path=True,
            help="Generate a semantic version based on changes that impact the specified path.",
        ),
    ] = Path.cwd(),
    style: Annotated[
        GenerateStyle,
        typer.Option(
            "--style",
            case_sensitive=False,
            help="Specifies the way in which the semantic version is generated; this is useful when targets using the generated semantic version do not fully support the semantic version specification.",
        ),
    ] = GenerateStyle.Standard,
    prerelease_name: Annotated[
        Optional[str],
        typer.Option(
            "--prerelease-name",
            help="Create a semantic version string with this prerelease name.",
        ),
    ] = None,
    no_prefix: Annotated[
        bool,
        typer.Option(
            "--no-prefix",
            help="Do not include the prefix in the generated semantic version.",
        ),
    ] = False,
    no_branch_name: Annotated[
        bool,
        typer.Option(
            "--no-branch-name",
            help="Do not include the branch name in the prerelease section of the generated semantic version.",
        ),
    ] = False,
    no_metadata: Annotated[
        bool,
        typer.Option(
            "--no-metadata",
            help="Do not include the build metadata section of the generated semantic version.",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose",
            help="Write verbose information to the terminal.",
        ),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            help="Write debug information to the terminal.",
        ),
    ] = False,
    quiet: Annotated[
        bool,
        typer.Option(
            "--quiet",
            help="Do not display any information other than the generated semantic version.",
        ),
    ] = False,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            help="Display the version of this tool and exit.",
        ),
    ] = False,
) -> None:
    if version:
        sys.stdout.write("autogitsemver v{}\n".format(__version__))
        sys.exit(0)

    output_stream: Optional[TextWriterT] = None
    postprocess_func: Optional[Callable[[DoneManager, Optional[GetSemanticVersionResult]], None]] = None

    if quiet:
        sink = StringIO()

        output_stream = sink

        # ----------------------------------------------------------------------
        def PostprocessQuietData(
            dm: DoneManager,
            result: Optional[GetSemanticVersionResult],
        ) -> None:
            if dm.result != 0:
                sys.stdout.write(sink.getvalue())
            else:
                assert result is not None
                sys.stdout.write(result.semantic_version_string)

        # ----------------------------------------------------------------------

        postprocess_func = PostprocessQuietData

    else:
        output_stream = sys.stdout

        # ----------------------------------------------------------------------
        def PostprocessNone(*args, **kwargs):
            return None

        # ----------------------------------------------------------------------

        postprocess_func = PostprocessNone

    assert output_stream is not None
    assert postprocess_func is not None

    with DoneManager.CreateCommandLine(
        output_stream,
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        result: Optional[GetSemanticVersionResult] = None

        with ExitStack(lambda: postprocess_func(dm, result)):
            result = GetSemanticVersion(
                dm,
                path,
                prerelease_name=prerelease_name,
                include_branch_name_when_necessary=not no_branch_name,
                no_prefix=no_prefix,
                no_metadata=no_metadata,
                style=style,
            )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
if __name__ == "__main__":
    app()



# ----------------------------------------------------------------------
# |
# |  EntryPoint_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-03-02 11:38:32
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
# """Unit tests for EntryPoint.py."""

from pathlib import Path
from typing import Any, Mapping
from unittest.mock import MagicMock as Mock, patch

from typer.testing import CliRunner

from AutoGitSemVer import GenerateStyle, GetSemanticVersionResult
from AutoGitSemVer.EntryPoint import app


# ----------------------------------------------------------------------
def test_Default():
    output, args, kwargs = _Execute()

    assert len(args) == 2
    assert args[1] == Path.cwd()

    assert len(kwargs) == 5
    assert kwargs["prerelease_name"] is None
    assert kwargs["include_branch_name_when_necessary"] is True
    assert kwargs["no_prefix"] is False
    assert kwargs["no_metadata"] is False
    assert kwargs["style"] == GenerateStyle.Standard


# ----------------------------------------------------------------------
def test_CustomPath():
    custom_path = Path(__file__).parent.parent

    output, args, kwargs = _Execute(str(custom_path))

    assert len(args) == 2
    assert args[1] == custom_path


# ----------------------------------------------------------------------
def test_NonBoolArgs():
    output, args, kwargs = _Execute(
        "--prerelease-name",
        "prerelease_name",
        "--style",
        "AllPrerelease",
    )

    assert len(args) == 2
    assert args[1] == Path.cwd()

    assert len(kwargs) == 5
    assert kwargs["prerelease_name"] == "prerelease_name"
    assert kwargs["include_branch_name_when_necessary"] is True
    assert kwargs["no_prefix"] is False
    assert kwargs["no_metadata"] is False
    assert kwargs["style"] == GenerateStyle.AllPrerelease


# ----------------------------------------------------------------------
def test_BoolIncludeBranchNameWhenNecessary():
    output, args, kwargs = _Execute(
        "--no-branch-name",
    )

    assert len(kwargs) == 5
    assert kwargs["prerelease_name"] is None
    assert kwargs["include_branch_name_when_necessary"] is False
    assert kwargs["no_prefix"] is False
    assert kwargs["no_metadata"] is False
    assert kwargs["style"] == GenerateStyle.Standard


# ----------------------------------------------------------------------
def test_BoolNoPrefix():
    output, args, kwargs = _Execute(
        "--no-prefix",
    )

    assert len(kwargs) == 5
    assert kwargs["prerelease_name"] is None
    assert kwargs["include_branch_name_when_necessary"] is True
    assert kwargs["no_prefix"] is True
    assert kwargs["no_metadata"] is False
    assert kwargs["style"] == GenerateStyle.Standard


# ----------------------------------------------------------------------
def test_BoolNoMetadata():
    output, args, kwargs = _Execute(
        "--no-metadata",
    )

    assert len(kwargs) == 5
    assert kwargs["prerelease_name"] is None
    assert kwargs["include_branch_name_when_necessary"] is True
    assert kwargs["no_prefix"] is False
    assert kwargs["no_metadata"] is True
    assert kwargs["style"] == GenerateStyle.Standard


# ----------------------------------------------------------------------
def test_Quiet():
    output, args, kwargs = _Execute()
    assert output != "1.2.3"

    output, args, kwargs = _Execute("--quiet")
    assert output == "1.2.3"


# ----------------------------------------------------------------------
def test_Version():
    output, args, kwargs = _Execute("--version")
    assert output.startswith("autogitsemver v")

    assert not args
    assert not kwargs


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _Execute(
    *args,
) -> tuple[str, tuple[Any, ...], Mapping[str, Any]]:
    with patch(
        "AutoGitSemVer.EntryPoint.GetSemanticVersion",
        return_value=GetSemanticVersionResult(None, Mock(), "1.2.3"),
    ) as mock:
        result = CliRunner().invoke(app, list(args))
        assert result.exit_code == 0

        if not mock.call_args_list:
            return result.output, (), {}

        assert len(mock.call_args_list) == 1
        return result.output, mock.call_args_list[0].args, mock.call_args_list[0].kwargs



# ----------------------------------------------------------------------
# |
# |  Lib.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-21 17:14:17
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Contains functionality used to generate a semantic version based on recent changes in an active git repository."""

import itertools
import json
import os
import platform
import re

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path, PurePath
from typing import Any, Callable, cast, ClassVar, Generator, Optional

import git
import rtyaml  # type: ignore [import-untyped]

from dbrownell_Common.InflectEx import inflect  # type: ignore [import-untyped]
from dbrownell_Common import PathEx  # type: ignore [import-untyped]
from dbrownell_Common.Streams.DoneManager import DoneManager  # type: ignore [import-untyped]
from jsonschema import Draft202012Validator, validators  # type: ignore [import-untyped]
from semantic_version import Version as SemVer  # type: ignore [import-untyped]


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
DEFAULT_CONFIGURATION_FILENAMES: list[str] = [
    "AutoGitSemVer.json",
    "AutoGitSemVer.yaml",
    "AutoGitSemVer.yml",
]


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class Configuration:
    """Data used to configure how the semantic version is generated."""

    # pylint: disable=too-many-instance-attributes

    # ----------------------------------------------------------------------
    filename: Optional[Path]

    version_prefix: Optional[str]

    prerelease_environment_variable_name: str

    initial_version: SemVer
    main_branch_names: list[str]

    additional_dependencies: list[Path]

    include_branch_name_when_necessary: bool = field(kw_only=True)
    include_timestamp_when_necessary: bool = field(kw_only=True)
    include_computer_name_when_necessary: bool = field(kw_only=True)


# ----------------------------------------------------------------------
class GenerateStyle(str, Enum):
    """Specifies the style by which the semantic version is generated."""

    Standard = "Standard"  # Uses both prerelease and build metadata: "1.2.3-prerelease+METADATA"
    AllPrerelease = "AllPrerelease"  # Combines metadata with prerelease data: "1.2.3-prerelease.METADATA"
    AllMetadata = "AllMetadata"  # Combines prerelease data with metadata: "1.2.3+prerelease.METADATA"


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class GetSemanticVersionResult:
    """Result of GetSemanticVersion."""

    configuration_filename: Optional[Path]
    semantic_version: SemVer
    semantic_version_string: str


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class CommitInfo:
    """Information about a commit."""

    # ----------------------------------------------------------------------
    WORKING_CHANGES_COMMIT_ID: ClassVar[str] = "<working>"

    id: str
    description: str
    tags: list[str]

    author: str
    author_date: datetime

    files: list[PurePath]


# ----------------------------------------------------------------------
@dataclass(frozen=True)
class VersionDelta:
    """Information about a delta applied to a semantic version based on information found in a commit."""

    # ----------------------------------------------------------------------
    major: int
    minor: int
    patch: int

    prerelease: Optional[str]
    build_metadata: Optional[str]

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return "{}.{}.{}{}{}".format(
            self.major,
            self.minor,
            self.patch,
            "-{}".format(self.prerelease) if self.prerelease else "",
            "+{}".format(self.build_metadata) if self.build_metadata else "",
        )


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def DefaultCommitDataExtractor(
    dm: DoneManager,  # pylint: disable=unused-argument
    commit_info: CommitInfo,
) -> Optional[VersionDelta]:
    """Default function used to extract semantic version deltas from git commits."""

    if "+major" in commit_info.description:
        return VersionDelta(1, 0, 0, None, None)
    if "+minor" in commit_info.description:
        return VersionDelta(0, 1, 0, None, None)
    if "+patch" in commit_info.description:
        return VersionDelta(0, 0, 1, None, None)
    if "+feature" in commit_info.description:
        return VersionDelta(0, 1, 0, None, None)

    return VersionDelta(0, 0, 1, None, None)


# ----------------------------------------------------------------------
def GetSemanticVersion(
    dm: DoneManager,
    path: Path,
    *,
    prerelease_name: Optional[str] = None,
    include_branch_name_when_necessary: bool = True,
    include_timestamp_when_necessary: bool = True,
    include_computer_name_when_necessary: bool = True,
    no_prefix: bool = False,
    no_metadata: bool = False,
    configuration_filenames: Optional[list[str]] = None,
    style: GenerateStyle = GenerateStyle.Standard,
    commit_delta_extraction_func: Callable[
        [
            DoneManager,
            CommitInfo,
        ],
        Optional[VersionDelta],  # None indicates that the commit does not impact the version
    ] = DefaultCommitDataExtractor,
) -> GetSemanticVersionResult:
    """Returns a semantic version based on git changes that impact the specified path."""

    configuration_filenames = configuration_filenames or DEFAULT_CONFIGURATION_FILENAMES

    repository_root = GetGitRoot(path)
    if repository_root is None:
        raise Exception("'{}' does not appear to be a git repository.".format(path))

    # Get the most applicable configuration
    configuration: Optional[Configuration] = None

    # ----------------------------------------------------------------------
    def DisplayConfiguration() -> str:
        if configuration is None:
            return "configuration errors were encountered"

        if configuration.filename is None:
            return "default configuration info will be used"

        return "configuration info found at '{}'".format(configuration.filename)

    # ----------------------------------------------------------------------

    with dm.Nested(
        "Loading AutoGitSemVer configuration...",
        DisplayConfiguration,
    ):
        configuration = GetConfiguration(path, configuration_filenames)

    changes_processed: int = 0
    version_deltas: list[VersionDelta] = []

    with dm.Nested(
        "Enumerating changes...",
        [
            lambda: "{} processed".format(inflect.no("change", changes_processed)),
            lambda: "{} applied [{:.02f}%]".format(
                inflect.no("change", len(version_deltas)),
                0 if changes_processed == 0 else ((len(version_deltas) / changes_processed) * 100),
            ),
        ],
    ) as enumerate_dm:
        root_path = configuration.filename.parent if configuration.filename else repository_root

        additional_dependency_lookup: set[Path] = set()

        for additional_dependency in configuration.additional_dependencies:
            if additional_dependency.is_file():
                additional_dependency_lookup.add(additional_dependency)
                continue

            if additional_dependency.is_dir():
                for ad_root, _, ad_filenames in os.walk(additional_dependency):
                    ad_root_path = Path(ad_root)

                    for ad_filename in ad_filenames:
                        additional_dependency_lookup.add(ad_root_path / ad_filename)

                continue

            assert False, additional_dependency  # pragma: no cover

        # ----------------------------------------------------------------------
        def IsAdditionalDependency(
            filename: Path,
        ) -> bool:
            return filename in additional_dependency_lookup

        # ----------------------------------------------------------------------
        def GetConfigurationPathForFile(
            filename: Path,
        ) -> Path:
            result = GetConfigurationFilename(filename.parent, configuration_filenames)

            if result is not None:
                return result.parent

            return repository_root

        # ----------------------------------------------------------------------
        def ShouldProcess(
            commit: CommitInfo,
        ) -> bool:
            for filename in commit.files:
                fullpath = repository_root / filename

                if (
                    PathEx.IsDescendant(fullpath, root_path)
                    and GetConfigurationPathForFile(fullpath) == root_path
                ) or IsAdditionalDependency(fullpath):
                    return True

            return False

        # ----------------------------------------------------------------------

        version_regex_str = r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:-(?P<prerelease>[^\+]+))?(?:\+(?P<metadata>.+))?"

        if configuration.version_prefix:
            version_regex_str = r"^{}{}{}$".format(
                re.escape(configuration.version_prefix),
                "" if configuration.version_prefix.endswith("v") else "v?",
                version_regex_str,
            )
        else:
            version_regex_str = r"^v?{}$".format(version_regex_str)

        version_regex = re.compile(version_regex_str)

        # ----------------------------------------------------------------------
        def ExtractVersionFromTags(
            tags: list[str],
        ) -> Optional[VersionDelta]:
            for tag in tags:
                match = version_regex.search(tag)
                if match is None:
                    continue

                return VersionDelta(
                    int(match.group("major")),
                    int(match.group("minor")),
                    int(match.group("patch")),
                    match.group("prerelease"),
                    match.group("metadata"),
                )

            return None

        # ----------------------------------------------------------------------

        repo = git.Repo(repository_root)

        initial_version = VersionDelta(
            configuration.initial_version.major or 0,
            configuration.initial_version.minor or 0,
            configuration.initial_version.patch or 0,
            None,
            None,
        )

        for commit in EnumCommits(repo):
            changes_processed += 1

            if not ShouldProcess(commit):
                continue

            delta_applied: Optional[VersionDelta] = None

            with enumerate_dm.VerboseNested(
                "Processing '{}' ({})".format(commit.id, commit.author_date),
                lambda: str(delta_applied) if delta_applied else None,
            ):
                delta_applied = ExtractVersionFromTags(commit.tags)
                if delta_applied is not None:
                    initial_version = delta_applied
                    break

                delta_applied = commit_delta_extraction_func(enumerate_dm, commit)
                if delta_applied is None:
                    continue

                version_deltas.append(delta_applied)

    with dm.Nested("Calculating semantic version...") as calculate_dm:
        major = initial_version.major
        minor = initial_version.minor
        patch = initial_version.patch

        prerelease: list[str] = list(initial_version.prerelease or [])
        metadata: list[str] = list(initial_version.build_metadata or [])

        for version_delta in reversed(version_deltas):
            if version_delta.major:
                major += version_delta.major
                minor = 0
                patch = 0

                prerelease = []
                metadata = []

            if version_delta.minor:
                minor += version_delta.minor
                patch = 0

                prerelease = []
                metadata = []

            if version_delta.patch:
                patch += version_delta.patch

                prerelease = []
                metadata = []

            if version_delta.prerelease:
                prerelease.append(version_delta.prerelease)
            if version_delta.build_metadata:
                metadata.append(version_delta.build_metadata)

        if major == 0 and minor == 0:
            # If here, we are going to bump the minor version, so subtract a value from the patch value
            minor = 1

            if patch > 0:
                patch -= 1

        # Augment the prerelease items (if necessary)
        augmented_prerelease: list[str] = []

        prerelease_name = prerelease_name or os.getenv(  # pylint: disable=invalid-envvar-value
            configuration.prerelease_environment_variable_name
        )
        if prerelease_name is not None:
            augmented_prerelease.append(prerelease_name)

        # ----------------------------------------------------------------------
        def GetBranchName() -> str:
            try:
                return repo.active_branch.name
            except TypeError:
                return "<detached head>"

        # ----------------------------------------------------------------------

        branch_name = GetBranchName()

        if (
            configuration.include_branch_name_when_necessary
            and include_branch_name_when_necessary
            and branch_name not in configuration.main_branch_names
        ):
            augmented_prerelease.append(branch_name)

        # Augment the metadata items (if necessary)
        augmented_metadata: list[str] = []

        if configuration.include_timestamp_when_necessary and include_timestamp_when_necessary:
            now = datetime.now()

            augmented_metadata.append(
                "{:04d}{:02d}{:02d}{:02d}{:02d}{:02d}".format(
                    now.year,
                    now.month,
                    now.day,
                    now.hour,
                    now.minute,
                    now.second,
                ),
            )

        if configuration.include_computer_name_when_necessary and include_computer_name_when_necessary:
            augmented_metadata.append(platform.node())

        if repo.is_dirty():
            augmented_metadata.append("working_changes")

        # Combine the elements
        prerelease = augmented_prerelease + prerelease
        metadata = augmented_metadata + metadata

        if style == GenerateStyle.Standard:
            # No changes are necessary
            pass
        elif style == GenerateStyle.AllPrerelease:
            prerelease += metadata
            metadata = []
        elif style == GenerateStyle.AllMetadata:
            metadata = prerelease + metadata
            prerelease = []
        else:
            assert False, style  # pragma: no cover

        # Create the semantic version
        semver = SemVer(
            major=major,
            minor=minor,
            patch=patch,
            prerelease=(None if no_prefix else tuple(prerelease)),
            build=None if no_metadata else tuple(metadata),
        )

        semver_string = f"{configuration.version_prefix or ''}{semver}"

        calculate_dm.WriteLine(semver_string)

    return GetSemanticVersionResult(configuration.filename, semver, semver_string)


# ----------------------------------------------------------------------
def GetConfiguration(
    path: Path,
    configuration_filenames: Optional[list[str]] = None,
) -> Configuration:
    """Returns the configuration data impacting the specified path."""

    # Get the configuration filename
    configuration_filename: Optional[Path] = GetConfigurationFilename(path, configuration_filenames)

    # Get the configuration content
    configuration_content: dict[str, Any] = {}

    if configuration_filename is not None:
        with configuration_filename.open() as f:
            if configuration_filename.suffix in [".yaml", ".yml"]:
                configuration_content = cast(dict[str, Any], rtyaml.load(f))
            elif configuration_filename.suffix == ".json":
                configuration_content = json.load(f)
            else:
                assert False, configuration_filename  # pragma: no cover

    # Load the schema
    schema_filename = Path(__file__).parent / "AutoGitSemVerSchema.json"
    if not schema_filename.is_file():
        raise Exception("The filename '{}' does not exist.".format(schema_filename))  # pragma: no cover

    with schema_filename.open() as f:
        schema_content = json.load(f)

    # Create the configuration validator. The special class is augmented to apply defaults to the
    # configuration. This code is based on https://python-jsonschema.readthedocs.io/en/latest/faq/
    validator_class = Draft202012Validator
    validate_properties = validator_class.VALIDATORS["properties"]  # type: ignore

    # ----------------------------------------------------------------------
    def SetDefaults(validator, properties, instance, schema):
        for prop, sub_schema in properties.items():
            default_schema = sub_schema.get("default", None)
            if default_schema is not None:
                instance.setdefault(prop, default_schema)

            for error in validate_properties(validator, properties, instance, schema):
                yield error

    # ----------------------------------------------------------------------

    validator = validators.extend(validator_class, {"properties": SetDefaults})(schema_content)

    # Validate the configuration data
    validator.validate(configuration_content)

    additional_dependencies: list[Path] = []

    if configuration_filename is not None:
        for additional_dependency in configuration_content.get("additional_dependencies", []):
            fullpath = (configuration_filename.parent / additional_dependency).resolve()

            if not fullpath.exists():
                raise Exception("The additional dependency '{}' does not exist.".format(fullpath))

            additional_dependencies.append(fullpath)

    return Configuration(
        configuration_filename,
        configuration_content.get("version_prefix", None),
        configuration_content["prerelease_environment_variable_name"],
        SemVer.coerce(configuration_content["initial_version"]),
        configuration_content["main_branch_names"],
        additional_dependencies,
        include_branch_name_when_necessary=configuration_content["include_branch_name_when_necessary"],
        include_timestamp_when_necessary=configuration_content["include_timestamp_when_necessary"],
        include_computer_name_when_necessary=configuration_content["include_computer_name_when_necessary"],
    )


# ----------------------------------------------------------------------
def GetConfigurationFilename(
    path: Path,
    configuration_filenames: Optional[list[str]] = None,
) -> Optional[Path]:
    """Returns the configuration filename impacting the specified path."""

    configuration_filenames = configuration_filenames or DEFAULT_CONFIGURATION_FILENAMES

    for parent in itertools.chain([path], path.parents):
        for potential_configuration_filename in configuration_filenames:
            potential_filename = parent / potential_configuration_filename
            if potential_filename.is_file():
                return potential_filename

        if (parent / ".git").is_dir():
            break

    return None


# ----------------------------------------------------------------------
def GetGitRoot(path: Path) -> Optional[Path]:
    """Returns the root of the git repository associated with the provided path."""

    for root in itertools.chain([path], path.parents):
        if (root / ".git").is_dir():
            return root

    return None


# ----------------------------------------------------------------------
def EnumCommits(
    repo_or_path: git.Repo | Path,
) -> Generator[CommitInfo, None, None]:
    """Enumerates git commits for the specified repository and branch."""

    if isinstance(repo_or_path, Path):
        repo = git.Repo(repo_or_path)
    elif isinstance(repo_or_path, git.Repo):
        repo = repo_or_path
    else:
        assert False, repo_or_path  # pragma: no cover

    # Return the working changes (if any)
    if repo.is_dirty():
        yield CommitInfo(
            CommitInfo.WORKING_CHANGES_COMMIT_ID,
            "",
            [],
            "",
            datetime.now(),
            [PurePath(diff.a_path) for diff in repo.active_branch.commit.diff()],
        )

    # Enumerate commits
    git_tags: dict[str, list[git.Tag]] = {}

    for tag in repo.tags:
        # Tags are most often associated with merges into a mainline branch, but we are filtering merges
        # out in the code below. Therefore, associate the tag with a parent that isn't a merge commit.
        if len(tag.commit.parents) == 1:
            # We are looking at a direct commit to the branch
            git_tags.setdefault(tag.commit.hexsha, []).append(tag)
        else:
            # We are looking at a merge
            for parent in tag.commit.parents:
                if len(parent.parents) == 1:
                    git_tags.setdefault(parent.hexsha, []).append(tag)
                    break

    offset = 0

    while True:
        commits = list(repo.iter_commits("HEAD", max_count=50, skip=offset))
        if not commits:
            break

        offset += len(commits)

        for commit in commits:
            # Ignore merge commits
            if len(commit.parents) > 1:
                continue

            assert isinstance(commit.message, str), commit.message

            yield CommitInfo(
                commit.hexsha,
                commit.message,
                [tag.name for tag in git_tags.get(commit.hexsha, [])],
                commit.author.name or "",
                commit.authored_datetime,
                [PurePath(path) for path in commit.stats.files],
            )



# noqa: D104

# Wheel names will be generated according to this value. Do not manually modify this value; instead
# update it according to committed changes by running this command from the root of the repository:
#
#   uv run python -m AutoGitSemVer.scripts.UpdatePythonVersion ./src/AutoGitSemVer/__init__.py ./src
#
__version__ = "0.9.1"


from .Lib import GenerateStyle, GetSemanticVersion, GetSemanticVersionResult

__all__ = [
    "GenerateStyle",
    "GetSemanticVersion",
    "GetSemanticVersionResult",
]



# ----------------------------------------------------------------------
# |
# |  Lib_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-26 07:26:21
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit test for AutoGitSemVer/Lib.py"""

import re
import textwrap

from io import StringIO
from unittest.mock import patch
from uuid import uuid4

import pytest

from dbrownell_Common import SubprocessEx  # type: ignore [import-untyped]

from AutoGitSemVer.Lib import *  # type: ignore [import-untyped]


# ----------------------------------------------------------------------
def test_EnumCommits(tmp_path_factory):
    # Create the scenario
    repo_dir = tmp_path_factory.mktemp("repo")

    assert SubprocessEx.Run("git init", cwd=repo_dir).returncode == 0
    assert SubprocessEx.Run('git config user.name "Test User"', cwd=repo_dir).returncode == 0
    assert SubprocessEx.Run('git config user.email "a@b.com"', cwd=repo_dir).returncode == 0

    # Commit 1
    for index in range(2):
        with (repo_dir / "FileA{}.txt".format(index)).open("w") as f:
            pass

    assert SubprocessEx.Run("git add .", cwd=repo_dir).returncode == 0
    assert SubprocessEx.Run('git commit -m "Commit 1"', cwd=repo_dir).returncode == 0

    # Commit 2
    for index in range(1):
        with (repo_dir / "FileB{}.txt".format(index)).open("w") as f:
            pass

    assert SubprocessEx.Run("git add .", cwd=repo_dir).returncode == 0
    assert SubprocessEx.Run('git commit -m "Commit 2"', cwd=repo_dir).returncode == 0

    # Working
    for index in range(4):
        with (repo_dir / "FileC{}.txt".format(index)).open("w") as f:
            pass

    # ----------------------------------------------------------------------
    def Validate(
        commits: list[CommitInfo],
    ) -> None:
        # Commit #2
        assert commits[-2].id != CommitInfo.WORKING_CHANGES_COMMIT_ID
        assert commits[-2].files == [
            PurePath("FileB0.txt"),
        ]

        # Commit #1
        assert commits[-1].id != CommitInfo.WORKING_CHANGES_COMMIT_ID
        assert commits[-1].files == [
            PurePath("FileA0.txt"),
            PurePath("FileA1.txt"),
        ]

    # ----------------------------------------------------------------------

    # Scenario #1: Working changes, but nothing staged
    commits = list(EnumCommits(repo_dir))

    Validate(commits)

    assert len(commits) == 2

    # Scenario #2: Some staged changes
    assert SubprocessEx.Run("git add FileC1.txt FileC3.txt", cwd=repo_dir).returncode == 0

    commits = list(EnumCommits(repo_dir))

    Validate(commits)

    assert len(commits) == 3
    assert commits[0].id == CommitInfo.WORKING_CHANGES_COMMIT_ID
    assert commits[0].files == [
        PurePath("FileC1.txt"),
        PurePath("FileC3.txt"),
    ]


# ----------------------------------------------------------------------
def test_GetGitRoot():
    this_dir = Path(__file__).parent
    assert GetGitRoot(this_dir) == this_dir.parent
    assert GetGitRoot(this_dir.parent.parent) is None


# ----------------------------------------------------------------------
def test_GetConfigurationFilename():
    this_dir = Path(__file__).parent

    assert GetConfigurationFilename(this_dir) is None
    assert (
        GetConfigurationFilename(this_dir.parent / "src" / "AutoGitSemVer")
        == this_dir.parent / "src" / "AutoGitSemVer.yaml"
    )


# ----------------------------------------------------------------------
class TestGetConfiguration:
    # ----------------------------------------------------------------------
    def test_Default(self):
        configuration = GetConfiguration(Path(__file__).parent)

        assert configuration.filename is None
        assert configuration.include_branch_name_when_necessary
        assert configuration.include_computer_name_when_necessary
        assert configuration.include_timestamp_when_necessary
        assert configuration.initial_version == SemVer("0.0.0")
        assert configuration.main_branch_names == ["main", "master", "default"]
        assert configuration.prerelease_environment_variable_name == "AUTO_GIT_SEM_VER_PRERELEASE_NAME"
        assert configuration.version_prefix is None
        assert configuration.additional_dependencies == []

    # ----------------------------------------------------------------------
    def test_Yaml(self, tmp_path_factory):
        root = tmp_path_factory.mktemp("root")

        with (root / "one").open("w") as f:
            pass

        with (root / "two").open("w") as f:
            pass

        with (root / "AutoGitSemVer.yaml").open("w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    {
                        include_branch_name_when_necessary: false,
                        include_computer_name_when_necessary: false,
                        include_timestamp_when_necessary: false,
                        initial_version: 1.2.3,
                        main_branch_names: [ "foo", "bar" ],
                        prerelease_environment_variable_name: "BAZ",
                        additional_dependencies: ["one", "two"]
                    }
                    """,
                ),
            )

        configuration = GetConfiguration(root)

        assert configuration.filename == root / "AutoGitSemVer.yaml"
        assert not configuration.include_branch_name_when_necessary
        assert not configuration.include_computer_name_when_necessary
        assert not configuration.include_timestamp_when_necessary
        assert configuration.initial_version == SemVer("1.2.3")
        assert configuration.main_branch_names == ["foo", "bar"]
        assert configuration.prerelease_environment_variable_name == "BAZ"
        assert configuration.version_prefix is None
        assert configuration.additional_dependencies == [root / "one", root / "two"]

    # ----------------------------------------------------------------------
    def test_Yml(self, tmp_path_factory):
        root = tmp_path_factory.mktemp("root")

        with (root / "AutoGitSemVer.yml").open("w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    {
                        include_timestamp_when_necessary: false,
                        initial_version: 1.2.3,
                        version_prefix: "Test-v"
                    }
                    """,
                ),
            )

        configuration = GetConfiguration(root)

        assert configuration.filename == root / "AutoGitSemVer.yml"
        assert configuration.include_branch_name_when_necessary
        assert configuration.include_computer_name_when_necessary
        assert not configuration.include_timestamp_when_necessary
        assert configuration.initial_version == SemVer("1.2.3")
        assert configuration.main_branch_names == ["main", "master", "default"]
        assert configuration.prerelease_environment_variable_name == "AUTO_GIT_SEM_VER_PRERELEASE_NAME"
        assert configuration.version_prefix == "Test-v"
        assert configuration.additional_dependencies == []

    # ----------------------------------------------------------------------
    def test_Json(self, tmp_path_factory):
        root = tmp_path_factory.mktemp("root")

        with (root / "AutoGitSemVer.json").open("w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    {
                        "include_branch_name_when_necessary": false
                    }
                    """,
                ),
            )

        configuration = GetConfiguration(root)

        assert configuration.filename == root / "AutoGitSemVer.json"
        assert not configuration.include_branch_name_when_necessary
        assert configuration.include_computer_name_when_necessary
        assert configuration.include_timestamp_when_necessary
        assert configuration.initial_version == SemVer("0.0.0")
        assert configuration.main_branch_names == ["main", "master", "default"]
        assert configuration.prerelease_environment_variable_name == "AUTO_GIT_SEM_VER_PRERELEASE_NAME"
        assert configuration.version_prefix is None
        assert configuration.additional_dependencies == []

    # ----------------------------------------------------------------------
    def test_Invalid(self, tmp_path_factory):
        root = tmp_path_factory.mktemp("root")

        with (root / "AutoGitSemVer.yaml").open("w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    {
                        include_branch_name_when_necessary: "This is not a valid value"
                    }
                    """,
                ),
            )

        with pytest.raises(
            Exception,
            match=re.compile("'This is not a valid value' is not of type 'boolean'.*", re.DOTALL),
        ):
            GetConfiguration(root)


# ----------------------------------------------------------------------
class TestSemanticVersion:
    # ----------------------------------------------------------------------
    def test_Empty(self):
        result, semver = _GetSemanticVersionImpl([])

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0

    # ----------------------------------------------------------------------
    def test_Single(self):
        result, semver = _GetSemanticVersionImpl([_CreateCommitInfo("Commit 1")])

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0

    # ----------------------------------------------------------------------
    def test_Multiple(self):
        commits = [
            _CreateCommitInfo("Commit 3"),
            _CreateCommitInfo("Commit 2"),
            _CreateCommitInfo("Commit 1"),
        ]

        result, semver = _GetSemanticVersionImpl(commits)

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == len(commits) - 1

    # ----------------------------------------------------------------------
    def test_BumpMinor(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("Bump Patch 3"),
                _CreateCommitInfo("Bump Minor (+minor)"),  # This will reset the patch
                _CreateCommitInfo("Bump Patch 2"),
                _CreateCommitInfo("Bump Patch 1"),
            ],
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 1

    # ----------------------------------------------------------------------
    def test_BumpMajor(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("Bump Patch (+patch)"),
                _CreateCommitInfo("Bump Minor (+minor)"),
                _CreateCommitInfo("Bump Major (+major)"),  # This will reset the minor
                _CreateCommitInfo("Bump Minor (+minor)"),
                _CreateCommitInfo("Bump Minor (+minor)"),
                _CreateCommitInfo("Bump Minor (+minor)"),
            ],
        )

        assert result == 0
        assert semver.semantic_version.major == 1
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 1

    # ----------------------------------------------------------------------
    def test_Feature(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("Feature 1 (+features)"),
                _CreateCommitInfo("Feature 2 (+feature)"),
                _CreateCommitInfo("Feature 3 (+features)"),
            ],
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 3
        assert semver.semantic_version.patch == 0

    # ----------------------------------------------------------------------
    def test_WithTag(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("(+major)"),
                _CreateCommitInfo("(+minor)"),
                _CreateCommitInfo("(+patch)"),
                _CreateCommitInfo("Tag", tags=["Ignore Me", "1.2.3"]),
            ],
        )

        assert result == 0
        assert semver.semantic_version.major == 2
        assert semver.semantic_version.minor == 0
        assert semver.semantic_version.patch == 0

    # ----------------------------------------------------------------------
    def test_WithTag2(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo(""),
                _CreateCommitInfo("", tags=["1.2.3"]),
            ],
        )

        assert result == 0
        assert semver.semantic_version.major == 1
        assert semver.semantic_version.minor == 2
        assert semver.semantic_version.patch == 4

    # ----------------------------------------------------------------------
    def test_PrereleaseName(self):
        result, semver = _GetSemanticVersionImpl([], prerelease_name="MyPrereleaseName")

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0
        assert (
            semver.semantic_version.prerelease and semver.semantic_version.prerelease[0] == "MyPrereleaseName"
        )

    # ----------------------------------------------------------------------
    def test_IgnoredChanges(self):
        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("Ignored 1", files=[PurePath("Ignored.txt")]),
                _CreateCommitInfo("Included 1", files=[PurePath("src/Included.txt")]),
                _CreateCommitInfo("Ignored 2", files=[PurePath("Ignored.txt")]),
                _CreateCommitInfo("Included 2", files=[PurePath("src/Included.txt")]),
                _CreateCommitInfo("Ignored 3", files=[PurePath("Ignored.txt")]),
                _CreateCommitInfo("Included 3", files=[PurePath("src/Included.txt")]),
            ],
            working_dir=Path(__file__).parent.parent / "src",
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 2

    # ----------------------------------------------------------------------
    def test_Styles(self):
        # Standard
        result, semver = _GetSemanticVersionImpl(
            [],
            prerelease_name="MyPrereleaseName",
            style=GenerateStyle.Standard,
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0

        assert semver.semantic_version.prerelease is not None and len(semver.semantic_version.prerelease) >= 1
        standard_prerelease = semver.semantic_version.prerelease

        assert semver.semantic_version.build is not None and len(semver.semantic_version.build) >= 1
        standard_build = semver.semantic_version.build

        # AllPrerelease
        result, semver = _GetSemanticVersionImpl(
            [],
            prerelease_name="MyPrereleaseName",
            style=GenerateStyle.AllPrerelease,
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0

        assert semver.semantic_version.prerelease is not None and (
            len(semver.semantic_version.prerelease) == len(standard_prerelease) + len(standard_build)
        )
        assert semver.semantic_version.build == ()

        # AllMetadata
        result, semver = _GetSemanticVersionImpl(
            [],
            prerelease_name="MyPrereleaseName",
            style=GenerateStyle.AllMetadata,
        )

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0

        assert semver.semantic_version.prerelease == ()
        assert semver.semantic_version.build is not None and (
            len(semver.semantic_version.build) == len(standard_prerelease) + len(standard_build)
        )

    # ----------------------------------------------------------------------
    def test_CustomDeltaFunc(self):
        # ----------------------------------------------------------------------
        def CustomCommitDeltaExtractor(
            dm: DoneManager,
            commit_info: CommitInfo,
        ) -> Optional[VersionDelta]:
            if "MAJOR" in commit_info.description:
                return VersionDelta(1000, 0, 0, "bumped_major", None)
            if "MINOR" in commit_info.description:
                return VersionDelta(0, 100, 0, "bumped_minor", None)
            if "PATCH" in commit_info.description:
                return VersionDelta(0, 0, 10, None, "bumped_patch")

            return None

        result, semver = _GetSemanticVersionImpl(
            [
                _CreateCommitInfo("PATCH"),
                _CreateCommitInfo(""),
                _CreateCommitInfo("MINOR"),
                _CreateCommitInfo(""),
                _CreateCommitInfo("MAJOR"),
                _CreateCommitInfo(""),
            ],
            commit_delta_extraction_func=CustomCommitDeltaExtractor,
        )

        assert result == 0
        assert semver.semantic_version.major == 1000
        assert semver.semantic_version.minor == 100
        assert semver.semantic_version.patch == 10
        assert (
            semver.semantic_version.prerelease is None
            or "bumped_major" not in semver.semantic_version.prerelease
        )  # prerelease is wiped when a new version is encountered
        assert (
            semver.semantic_version.prerelease is None
            or "bumped_minor" not in semver.semantic_version.prerelease
        )  # prerelease is wiped when a new version is encountered
        assert semver.semantic_version.build and "bumped_patch" in semver.semantic_version.build

    # ----------------------------------------------------------------------
    def test_String(self):
        result, semver = _GetSemanticVersionImpl([])

        assert result == 0
        assert semver.semantic_version.major == 0
        assert semver.semantic_version.minor == 1
        assert semver.semantic_version.patch == 0
        assert semver.semantic_version_string != "0.1.0"

        result, semver = _GetSemanticVersionImpl(
            [],
            include_branch_name_when_necessary=False,
            include_timestamp_when_necessary=False,
            include_computer_name_when_necessary=False,
            no_metadata=True,
        )
        assert semver.semantic_version_string == "0.1.0"


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _CreateCommitInfo(
    description: str,
    files: Optional[list[PurePath]] = None,
    tags: Optional[list[str]] = None,
    *,
    is_working_change: bool = False,
) -> CommitInfo:
    id = CommitInfo.WORKING_CHANGES_COMMIT_ID if is_working_change else str(uuid4())

    if files is None:
        files = [PurePath("File.txt")]

    return CommitInfo(id, description, tags or [], "Author", datetime.now(), files)


# ----------------------------------------------------------------------
def _GetSemanticVersionImpl(
    commits: list[CommitInfo],
    *,
    working_dir: Path = Path.cwd(),
    **kwargs,
) -> tuple[int, GetSemanticVersionResult]:
    with patch("AutoGitSemVer.Lib.EnumCommits", return_value=commits):
        sink = StringIO()

        with DoneManager.Create(sink, "_GetSemanticVersionImpl...") as dm:
            result = GetSemanticVersion(dm, working_dir, **kwargs)

        return dm.result, result



# ----------------------------------------------------------------------
# |
# |  UpdatePythonVersion.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2025-01-10 10:04:19
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2025
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Updates a __version__ variable with the calculated version in a Python file."""

import re

from pathlib import Path
from typing import Annotated, Optional

import typer

from AutoGitSemVer import GetSemanticVersion
from dbrownell_Common.Streams.DoneManager import DoneManager, Flags as DoneManagerFlags
from typer.core import TyperGroup


# ----------------------------------------------------------------------
class NaturalOrderGrouper(TyperGroup):
    # pylint: disable=missing-class-docstring
    # ----------------------------------------------------------------------
    def list_commands(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.commands.keys()  # pragma: no cover


# ----------------------------------------------------------------------
app = typer.Typer(
    cls=NaturalOrderGrouper,
    help=__doc__,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


# ----------------------------------------------------------------------
@app.command("EntryPoint", help=__doc__, no_args_is_help=True)
def EntryPoint(
    filename: Annotated[
        Path,
        typer.Argument(
            dir_okay=False,
            exists=True,
            resolve_path=True,
            help="Name of the python file that contains the __version__ variable or the pyproject.toml file that contains the version variable.",
        ),
    ],
    working_dir: Annotated[
        Optional[Path],
        typer.Argument(
            file_okay=False,
            exists=True,
            resolve_path=True,
            help="Working directory used to calculate the version.",
        ),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option("--verbose", help="Write verbose information to the terminal."),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option("--debug", help="Write debug information to the terminal."),
    ] = False,
) -> None:
    with DoneManager.CreateCommandLine(
        flags=DoneManagerFlags.Create(verbose=verbose, debug=debug),
    ) as dm:
        working_dir = working_dir or filename.parent

        version = GetSemanticVersion(
            dm,
            working_dir,
            no_metadata=True,
            include_branch_name_when_necessary=False,
            include_timestamp_when_necessary=False,
            include_computer_name_when_necessary=False,
        )

        dm.WriteLine("")

        with dm.Nested(f"Updating '{filename}'...") as update_dm:
            with filename.open(encoding="utf-8") as f:
                content = f.read()

            if filename.suffix == ".py":
                variable_name = "__version__"

                regex = re.compile(
                    rf"""
                    ^                       # Start of line
                    (?P<prefix>
                        \s*                 # Start of line and initial whitespace
                        {variable_name}     # Variable
                        \s*=\s*             # Equal
                        (?P<quote>          # Opening quote
                            '(?!')              # Single quote, not followed by another quote
                            |'''(?!')           # Triple quote, not followed by another quote
                            |\"(?!\")           # Double quote, not followed by another quote
                            |\"\"\"(?!\")       # Triple double quote, not followed by another quote
                        )
                    )
                    \S+?                    # Content
                    (?P<suffix>
                        (?P=quote)          # Closing quote
                        [^\n]*?             # Optional trailing content before newline
                    )
                    (?P<newline>\r?\n)      # Newline
                    """,
                    flags=re.MULTILINE | re.VERBOSE,
                )

            elif filename.suffix == ".toml":
                variable_name = "version"

                regex = re.compile(
                    rf"""
                    ^                       # Start of line
                    (?P<prefix>
                        \s*                 # Start of line and initial whitespace
                        {variable_name}     # Variable
                        \s*=\s*             # Equal
                        (?P<quote>          # Opening quote
                            \"(?!\")            # Double quote, not followed by another quote
                            |\"\"\"(?!\")       # Triple double quote, not followed by another quote
                        )
                    )
                    \S+?                    # Content
                    (?P<suffix>
                        (?P=quote)          # Closing quote
                        [^\n]*?             # Optional trailing content before newline
                    )
                    (?P<newline>\r?\n)      # Newline
                    """,
                    flags=re.MULTILINE | re.VERBOSE,
                )

            else:
                error = f"'{filename}' is not a recognized file type."
                raise Exception(error)

            match = regex.search(content)
            if not match:
                update_dm.WriteError(f"A '{variable_name}' variable was not found.\n")
                return

            content = "".join(
                [
                    content[: match.start()],
                    match.group("prefix"),
                    version.semantic_version_string,
                    match.group("suffix"),
                    match.group("newline"),
                    content[match.end() :],
                ],
            )

            with filename.open("w", encoding="utf-8", newline=match.group("newline")) as f:
                f.write(content)


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
if __name__ == "__main__":
    app()  # pragma: no cover



# ----------------------------------------------------------------------
# |
# |  UpdatePythonVersion_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2025-01-10 12:08:33
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2025
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for UpdatePythonVersion.py."""

import textwrap

from pathlib import Path
from unittest.mock import MagicMock as Mock, mock_open, patch

import pytest

from click.testing import Result
from typer.testing import CliRunner
from AutoGitSemVer import GetSemanticVersionResult
from AutoGitSemVer.scripts.UpdatePythonVersion import app


# ----------------------------------------------------------------------
@pytest.mark.parametrize("quote", ['"', "'", '"""', "'''"])
@pytest.mark.parametrize("prefix", ["", "    "])
@pytest.mark.parametrize("suffix", ["", "  # Comment!"])
def test_Python(quote, prefix, suffix):
    _Execute(
        textwrap.dedent(
            f"""\
            # Line Before
            {prefix}__version__ = {quote}0.0.0{quote}{suffix}
            # Line After
            """,
        ),
        textwrap.dedent(
            f"""\
            # Line Before
            {prefix}__version__ = {quote}1.2.3{quote}{suffix}
            # Line After
            """,
        ),
    )


# ----------------------------------------------------------------------
def test_NoPythonVersion():
    _Execute(
        textwrap.dedent(
            """\
            # No __version__ in this file
            """,
        ),
        "",
        expected_failure=True,
    )


# ----------------------------------------------------------------------
@pytest.mark.parametrize("quote", ['"', '"""'])
@pytest.mark.parametrize("prefix", ["", "    "])
@pytest.mark.parametrize("suffix", ["", "  # Comment!"])
def test_Toml(quote, prefix, suffix):
    _Execute(
        textwrap.dedent(
            f"""\
            # Line Before
            {prefix}version = {quote}0.0.0{quote}{suffix}
            # Line After
            """,
        ),
        textwrap.dedent(
            f"""\
            # Line Before
            {prefix}version = {quote}1.2.3{quote}{suffix}
            # Line After
            """,
        ),
        file_extension=".toml",
    )


# ----------------------------------------------------------------------
def test_NoTomlVersion():
    _Execute(
        textwrap.dedent(
            """\
            # No version in this file
            """,
        ),
        "",
        file_extension=".toml",
        expected_failure=True,
    )


# ----------------------------------------------------------------------
def test_UnsupportedFileType():
    result = _Execute(
        "",
        "",
        file_extension=".md",
        expected_failure=True,
    )
    exception_message = str(result.exception)
    assert "README.md' is not a recognized file type." in exception_message


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _Execute(
    original: str,
    expected: str,
    *,
    file_extension: str = ".py",
    expected_failure: bool = False,
) -> Result:
    # ----------------------------------------------------------------------
    def MyWrite(value):
        assert value == expected

    # ----------------------------------------------------------------------

    file_mock = mock_open(read_data=original)

    handle = file_mock()
    handle.read = Mock(return_value=original)
    handle.write = MyWrite

    with (
        patch(
            "AutoGitSemVer.scripts.UpdatePythonVersion.GetSemanticVersion",
            return_value=GetSemanticVersionResult(None, Mock(), "1.2.3"),
        ),
        patch.object(Path, "open", file_mock),
    ):
        if file_extension == ".py":
            placeholder_filename = Path(__file__)
        elif file_extension == ".toml":
            placeholder_filename = Path(__file__).parent.parent.parent / "pyproject.toml"
        elif file_extension == ".md":
            placeholder_filename = Path(__file__).parent.parent.parent / "README.md"
        else:
            assert False, file_extension

        result = CliRunner().invoke(app, [str(placeholder_filename)])

        if expected_failure:
            assert result.exit_code != 0, result.output
        else:
            assert result.exit_code == 0, result.output

        return result
